#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Drum model
The 1D drum model consists of three main unit operations
1) a flash model (only for water and steam)
2) a mixer model
3) a horizontal cylindrical water tank model
with multiple downcomers as water outlet

Inlet Ports:

* water/steam mixture from water wall
* feedwater_inlet: feedwater from economizer/pipe

Outlet Ports:

* liquid_outlet: liquid to downcomer
* steam_outlet: steam leaving the drum

The drum model receives saturated water from the water wall, and this is
separated at the flash unit, the steam leaves the drum, while the water/liquid
mixes with the feedwater from the economizer (or water pipe model).
Finally, the mixed state liquid (stream leaving the mixer), is used as the
control volume of the cylindrical water tank model,
which computes velocity and pressure
drop. The exit of the drum model is the liquid outlet.

main assumptions:

1.- Heat loss to ambient is calculated based on natural heat convection
2.- Calculate pressure change due to gravity based on water level
and contraction to downcomer
3.- Water level is either fixed for steady-state model or calculated for
dynamic model
4.- Assume enthalpy_in == enthalpy_out + heat loss
5.- Subcooled water from economizer and saturated water from waterwall
are mixed before entering drum
6.- Heat conduction through drum wall thickness is modeled
by 1D heat conduction

Created: October 27 2020
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import Pyomo libraries
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
from pyomo.environ import (
    value,
    Var,
    Param,
    Reference,
    exp,
    asin,
    cos,
    log10,
    log,
    sqrt,
    Constraint,
    TransformationFactory,
)
from pyomo.network import Port, Arc

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
import idaes.logger as idaeslog
from idaes.core.util.config import is_physical_parameter_block

# from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import fix_state_vars, revert_state_vars
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver

from idaes.models_extra.power_generation.unit_models.helm.phase_separator import (
    HelmPhaseSeparator,
)
from idaes.models_extra.power_generation.unit_models.helm import (
    HelmMixer,
    MomentumMixingType,
)
from idaes.core.util.constants import Constants as const

__author__ = "Boiler Subsystem Team (J. Ma, M. Zamarripa)"
__version__ = "2.0.0"


@declare_process_block_class("Drum1D")
class Drum1DData(UnitModelBlockData):
    """
    1-D Boiler Drum Unit Operation Class
    """

    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.componentPhase,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of material balance should be constructed,
**default** - MaterialBalanceType.componentPhase.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.enthalpyTotal,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single ethalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - ethalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_heat_transfer",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Heat transfer term construction flag",
            doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "finite_elements",
        ConfigValue(
            default=5,
            domain=int,
            description="Number of finite elements in length (drum radius) domain",
            doc="""Number of finite elements to use when discretizing length
domain (default=5). Should set to the number of tube rows""",
        ),
    )
    CONFIG.declare(
        "collocation_points",
        ConfigValue(
            default=3,
            domain=int,
            description="Number of collocation points per finite element",
            doc="""Number of collocation points to use per finite element when
discretizing length domain (default=3)""",
        ),
    )
    CONFIG.declare(
        "drum_inner_diameter",
        ConfigValue(
            default=1.0,
            description="inside diameter of drum",
            doc="define inside diameter of drum",
        ),
    )

    CONFIG.declare(
        "drum_thickness",
        ConfigValue(
            default=0.1,
            description="drum wall thickness",
            doc="define drum wall thickness",
        ),
    )

    def build(self):
        """
        Begin building model
        """

        # Call UnitModel.build to setup dynamics
        super().build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_geometry()

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type,
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer,
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type, has_pressure_change=True
        )

        self.flash = HelmPhaseSeparator(
            dynamic=False, property_package=self.config.property_package
        )

        self.mixer = HelmMixer(
            dynamic=False,
            property_package=self.config.property_package,
            momentum_mixing_type=MomentumMixingType.equality,
            inlet_list=["FeedWater", "SaturatedWater"],
        )

        # Inlet Ports
        # FeedWater to Drum (from Pipe or Economizer)
        self.feedwater_inlet = Port(extends=self.mixer.FeedWater)
        # Sat water from water wall
        self.water_steam_inlet = Port(extends=self.flash.inlet)

        # Exit Ports
        # Liquid to Downcomer
        # self.liquid_outlet = Port(extends=self.mixer.outlet)
        self.add_outlet_port("liquid_outlet", self.control_volume)
        # Steam to superheaters
        self.steam_outlet = Port(extends=self.flash.vap_outlet)

        self.stream_flash_out = Arc(
            source=self.flash.liq_outlet, destination=self.mixer.SaturatedWater
        )

        # Pyomo arc connect flash liq_outlet with mixer SaturatedWater inlet
        TransformationFactory("network.expand_arcs").apply_to(self)

        # connect internal units (Mixer to Water Tank Model)
        # Mixer Outlet (mixed_state) to unit control volume.properties_in
        @self.Constraint(self.flowsheet().time)
        def connection_material_balance(b, t):
            return (
                b.mixer.mixed_state[t].flow_mol
                == b.control_volume.properties_in[t].flow_mol
            )

        @self.Constraint(self.flowsheet().time)
        def connection_enthalpy_balance(b, t):
            return (
                b.mixer.mixed_state[t].enth_mol
                == b.control_volume.properties_in[t].enth_mol
            )

        @self.Constraint(self.flowsheet().time)
        def connection_pressure_balance(b, t):
            return (
                b.mixer.mixed_state[t].pressure
                == b.control_volume.properties_in[t].pressure
            )

        # Add object references
        self.volume = Reference(self.control_volume.volume)

        # Set references to balance terms at unit level
        if (
            self.config.has_heat_transfer is True
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            self.heat_duty = Reference(self.control_volume.heat)

        if (
            self.config.has_pressure_change is True
            and self.config.momentum_balance_type != "none"
        ):
            self.deltaP = Reference(self.control_volume.deltaP)

        # Set Unit Geometry and Holdup Volume
        self._set_geometry()

        # Construct performance equations
        self._make_performance()

    def _set_geometry(self):
        """
        Define the Geometry of the Unit
        """
        # Inside diameter of drum
        self.drum_diameter = Param(
            initialize=self.config.drum_inner_diameter, doc="Inside Diameter of Drum"
        )

        # Thickness of drum wall
        self.drum_thickness = Param(
            initialize=self.config.drum_thickness, doc="Wall Thickness of Drum"
        )

        # Thickness of insulation layer
        self.insulation_thickness = Var(
            initialize=0.2, doc="Insulation Layer Thickness"
        )

        # Length of drum
        self.drum_length = Var(initialize=10, doc="Horizontal Length of Drum")

        # Number of downcomers connected at the bottom of drum
        self.number_downcomer = Var(
            initialize=4, doc="Number of Downcomers Connected to Drum"
        )

        # Inside diameter of downcomer
        self.downcomer_diameter = Var(
            initialize=0.6, doc="Inside Diameter of Downcomer"
        )

        # Inside Radius expression
        @self.Expression(doc="Inside Radius of Drum")
        def drum_ri(b):
            return 0.5 * b.drum_diameter

        # Outside Radius expression
        @self.Expression(doc="Outside radius of drum")
        def drum_ro(b):
            return b.drum_ri + b.drum_thickness

        # Outside diameter expression
        @self.Expression(doc="Outside Radius of Drum")
        def drum_do(b):
            return b.drum_diameter + 2 * b.drum_thickness

        # Inner surface area (ignore two hemispheres at ends)
        @self.Expression(doc="Inner Surface Area")
        def drum_area(b):
            return const.pi * b.drum_diameter * b.drum_length

    def _make_performance(self):
        """
        Define constraints which describe the behaviour of the unit model.
        """

        # Thermal conductivity of drum
        self.therm_cond_metal = Param(initialize=40, mutable=True)

        # Thermal conductivity of insulation
        self.therm_cond_insulation = Param(initialize=0.08, mutable=True)

        # Thermal conductivity of air
        self.therm_cond_air = Param(initialize=0.03, mutable=True)

        # Density of drum material
        self.dens_metal = Param(initialize=7753, mutable=True)

        # Heat capacity of drum material
        self.cp_metal = Param(initialize=486, mutable=True)

        # Young modulus
        self.Young_modulus = Param(initialize=2.07e5, mutable=True)

        # Poisson's ratio
        self.Poisson_ratio = Param(initialize=0.292, mutable=True)

        # Coefficient of thermal expansion
        self.coefficient_therm_expansion = Param(initialize=1.4e-5, mutable=True)

        # Constant related to Rayleigh number for free convection
        # (gravity*expansion_coefficient*density
        # /viscosity/thermal_diffusivity)^(1/6)
        # Use properties at 50 C and 1 atm, 6.84e7^(1/6)=20.223
        self.const_Ra_root6 = Param(
            initialize=20.223, mutable=True, doc="Rayleigh number for free convection"
        )

        # Constant related to Nu for free convection,
        # 0.387/(1+0.721*Pr^(-9/16))^(8/27)
        # Use properties at 50 C and 1 atm
        self.const_Nu = Param(
            initialize=0.322,
            mutable=True,
            doc="constant related to Nu number for free" "convection",
        )

        # Ambient pressure
        self.pressure_amb = Param(initialize=1.0e5, doc="Ambient Pressure")

        # Ambient temperature
        self.temperature_ambient = Var(
            self.flowsheet().time, initialize=298.15, doc="Ambient Temperature"
        )

        # Inside heat transfer coefficient
        self.heat_transfer_in = Var(
            self.flowsheet().time, initialize=1, doc="Inside Heat Transfer Coefficient"
        )

        # Outside heat transfer coefficient
        self.heat_transfer_out = Var(
            self.flowsheet().time, initialize=1, doc="Outside Heat Transfer Coefficient"
        )

        # Insulation free convection heat transfer coefficient
        self.heat_transfer_free_conv = Var(
            self.flowsheet().time,
            initialize=1,
            doc="Insulation Free Convection" "Heat Transfer Coefficient",
        )

        # Ra number of free convection
        self.N_Ra_root6 = Var(
            self.flowsheet().time,
            initialize=80,
            doc="1/6 Power of Ra" "Number of Free Convection of Air",
        )

        # Nu number  of free convection
        self.N_Nu = Var(
            self.flowsheet().time,
            initialize=1,
            doc="Nu Number of Free Convection of Air",
        )

        # Define the continuous domains for model
        self.radial_domain = ContinuousSet(bounds=(self.drum_ri, self.drum_ro))

        # Temperature across wall thickness
        self.drum_wall_temperature = Var(
            self.flowsheet().time, self.radial_domain, bounds=(280, 800), initialize=550
        )

        # Declare derivatives in the model
        if self.config.dynamic is True:
            self.dTdt = DerivativeVar(
                self.drum_wall_temperature, wrt=self.flowsheet().time
            )
        self.dTdr = DerivativeVar(self.drum_wall_temperature, wrt=self.radial_domain)
        self.d2Tdr2 = DerivativeVar(
            self.drum_wall_temperature, wrt=(self.radial_domain, self.radial_domain)
        )

        discretizer = TransformationFactory("dae.finite_difference")
        discretizer.apply_to(
            self,
            nfe=self.config.finite_elements,
            wrt=self.radial_domain,
            scheme="CENTRAL",
        )

        # Add performance variables
        self.level = Var(
            self.flowsheet().time,
            initialize=1.0,
            doc="Water Level from the Bottom of the Drum",
        )

        # Velocity of fluid inside downcomer pipe
        self.velocity_downcomer = Var(
            self.flowsheet().time,
            initialize=10.0,
            doc="Liquid Water Velocity at the Top of Downcomer",
        )

        # Pressure change due to contraction
        self.deltaP_contraction = Var(
            self.flowsheet().time,
            initialize=-1.0,
            doc="Pressure Change due to Contraction",
        )

        # Pressure change due to gravity
        self.deltaP_gravity = Var(
            self.flowsheet().time, initialize=1.0, doc="Pressure Change due to Gravity"
        )

        # Thermal diffusivity of drum
        @self.Expression(doc="Thermal Diffusivity of Drum Material")
        def diff_therm_metal(b):
            return b.therm_cond_metal / (b.dens_metal * b.cp_metal)

        # Expressure for the angle from the drum center
        # to the circumference point at water level
        @self.Expression(self.flowsheet().time, doc="Angle of Water Level")
        def alpha_drum(b, t):
            return asin((b.level[t] - b.drum_ri) / b.drum_ri)

        # Expressure for the fraction of wet area
        @self.Expression(self.flowsheet().time, doc="Fraction of Wet Area")
        def frac_wet_area(b, t):
            return (b.alpha_drum[t] + const.pi / 2) / const.pi

        # Constraint for volume liquid in drum
        @self.Constraint(self.flowsheet().time, doc="Volume of Liquid in Drum")
        def volume_eqn(b, t):
            return (
                b.volume[t]
                == (
                    (b.alpha_drum[t] + 0.5 * const.pi) * b.drum_ri**2
                    + b.drum_ri * cos(b.alpha_drum[t]) * (b.level[t] - b.drum_ri)
                )
                * b.drum_length
            )

        # Equation for velocity at the entrance of downcomer
        @self.Constraint(self.flowsheet().time, doc="Velocity at Entrance of Downcomer")
        def velocity_eqn(b, t):
            return (
                b.velocity_downcomer[t]
                * 0.25
                * const.pi
                * b.downcomer_diameter**2
                * b.number_downcomer
                == b.control_volume.properties_out[t].flow_vol
            )

        # Pressure change equation for contraction,
        # -0.5*1/2*density*velocity^2 for stagnation head loss
        # plus 1/2*density*velocity^2 dynamic head
        # (acceleration pressure change)
        @self.Constraint(
            self.flowsheet().time, doc="Pressure Change due To Contraction"
        )
        def pressure_change_contraction_eqn(b, t):
            return (
                b.deltaP_contraction[t]
                == -0.75
                * b.control_volume.properties_out[t].dens_mass_phase["Liq"]
                * b.velocity_downcomer[t] ** 2
            )

        # Pressure change equation for gravity, density*gravity*height
        @self.Constraint(self.flowsheet().time, doc="Pressure Change due to Gravity")
        def pressure_change_gravity_eqn(b, t):
            return (
                b.deltaP_gravity[t]
                == b.control_volume.properties_out[t].dens_mass_phase["Liq"]
                * const.acceleration_gravity
                * b.level[t]
            )

        # Total pressure change equation
        @self.Constraint(self.flowsheet().time, doc="Pressure Drop")
        def pressure_change_total_eqn(b, t):
            return b.deltaP[t] == b.deltaP_contraction[t] + b.deltaP_gravity[t]

        # Constraint for heat conduction equation
        @self.Constraint(
            self.flowsheet().time,
            self.radial_domain,
            doc="1-D Heat Conduction Equation Through Radius",
        )
        def heat_conduction_eqn(b, t, r):
            if r == b.radial_domain.first() or r == b.radial_domain.last():
                return Constraint.Skip
            if self.config.dynamic is True:
                return (
                    b.dTdt[t, r]
                    == b.diff_therm_metal * b.d2Tdr2[t, r]
                    + b.diff_therm_metal * (1 / r) * b.dTdr[t, r]
                )
            else:
                return (
                    0
                    == b.diff_therm_metal * b.d2Tdr2[t, r]
                    + b.diff_therm_metal * (1 / r) * b.dTdr[t, r]
                )

        @self.Constraint(self.flowsheet().time, doc="Inner Wall Boundary")
        def inner_wall_bc_eqn(b, t):
            return (
                b.heat_transfer_in[t]
                * (
                    b.control_volume.properties_out[t].temperature
                    - b.drum_wall_temperature[t, b.radial_domain.first()]
                )
                == -b.dTdr[t, b.radial_domain.first()] * b.therm_cond_metal
            )

        @self.Constraint(self.flowsheet().time, doc="Outer Wall Boundary")
        def outer_wall_bc_eqn(b, t):
            return (
                b.heat_transfer_out[t]
                * (
                    b.drum_wall_temperature[t, b.radial_domain.last()]
                    - b.temperature_ambient[t]
                )
                == -b.dTdr[t, b.radial_domain.last()] * b.therm_cond_metal
            )

        # Inner wall BC for dTdt
        @self.Constraint(
            self.flowsheet().time, doc="Extra Inner Wall Temperature Derivative"
        )
        def extra_at_inner_wall_eqn(b, t):
            if self.config.dynamic is True:
                term = b.dTdt[t, b.radial_domain.first()]
            else:
                term = 0
            return term == 4 * b.diff_therm_metal * (
                b.radial_domain.first() + b.radial_domain.at(2)
            ) / (b.radial_domain.at(2) - b.radial_domain.first()) ** 2 / (
                3 * b.radial_domain.first() + b.radial_domain.at(2)
            ) * (
                b.drum_wall_temperature[t, b.radial_domain.at(2)]
                - b.drum_wall_temperature[t, b.radial_domain.first()]
            ) + 8 * b.diff_therm_metal / b.therm_cond_metal * b.heat_transfer_in[
                t
            ] * b.radial_domain.first() / (
                b.radial_domain.at(2) - b.radial_domain.first()
            ) / (
                3 * b.radial_domain.first() + b.radial_domain.at(2)
            ) * (
                b.control_volume.properties_out[t].temperature
                - b.drum_wall_temperature[t, b.radial_domain.first()]
            )

        @self.Constraint(
            self.flowsheet().time, doc="Extra Outer Wall Temperature Derivative"
        )
        def extra_at_outer_wall_eqn(b, t):
            if self.config.dynamic is True:
                term = b.dTdt[t, b.radial_domain.last()]
            else:
                term = 0
            return term == 4 * b.diff_therm_metal * (
                b.radial_domain.last() + b.radial_domain.at(-2)
            ) / (b.radial_domain.last() - b.radial_domain.at(-2)) ** 2 / (
                3 * b.radial_domain.last() + b.radial_domain.at(-2)
            ) * (
                b.drum_wall_temperature[t, b.radial_domain.at(-2)]
                - b.drum_wall_temperature[t, b.radial_domain.last()]
            ) + 8 * b.diff_therm_metal / b.therm_cond_metal * b.heat_transfer_out[
                t
            ] * b.radial_domain.last() / (
                b.radial_domain.last() - b.radial_domain.at(-2)
            ) / (
                3 * b.radial_domain.last() + b.radial_domain.at(-2)
            ) * (
                b.temperature_ambient[t]
                - b.drum_wall_temperature[t, b.radial_domain.last()]
            )

        # Reduced pressure expression
        @self.Expression(self.flowsheet().time, doc="Reduced Pressure")
        def pres_reduced(b, t):
            return b.control_volume.properties_out[t].pressure / 2.2048e7

        # Calculate inner side heat transfer coefficient
        # with minimum temperature difference set to sqrt(0.1)
        # multiplying wet area fraction to convert it
        # to the value based on total circumference
        @self.Constraint(
            self.flowsheet().time, doc="Inner Side Heat Transfer Coefficient"
        )
        def h_in_eqn(b, t):
            return (
                b.heat_transfer_in[t]
                == 2178.6
                * (b.control_volume.properties_out[t].pressure / 2.2048e7) ** 0.36
                / (-log10(b.control_volume.properties_out[t].pressure / 2.2048e7))
                ** 1.65
                * (
                    0.1
                    + (
                        b.control_volume.properties_out[t].temperature
                        - b.drum_wall_temperature[t, b.radial_domain.first()]
                    )
                    ** 2
                )
                * b.frac_wet_area[t]
            )

        # Expressure for insulation heat transfer (conduction)
        # resistance based on drum metal outside diameter
        @self.Expression(doc="Heat Transfer Resistance of Insulation Layer")
        def resistance_insulation(b):
            return (
                b.drum_ro
                * log((b.drum_ro + b.insulation_thickness) / b.drum_ro)
                / b.therm_cond_insulation
            )

        # heat_transfer_out equation considering conduction through insulation
        # and free convection between insulation and ambient
        @self.Constraint(
            self.flowsheet().time, doc="Outer Side Heat Transfer Coefficient"
        )
        def h_out_eqn(b, t):
            return (
                b.heat_transfer_out[t]
                * (b.resistance_insulation + 1 / b.heat_transfer_free_conv[t])
                == 1.0
            )

        # Expressure for outside insulation wall temperature (skin temperature)
        @self.Expression(
            self.flowsheet().time, doc="Outside Insulation Wall Temperature"
        )
        def temp_insulation_outside(b, t):
            return (
                b.temperature_ambient[t]
                + (
                    b.drum_wall_temperature[t, b.radial_domain.last()]
                    - b.temperature_ambient[t]
                )
                * b.heat_transfer_out[t]
                / b.heat_transfer_free_conv[t]
            )

        # Ra number equation
        @self.Constraint(self.flowsheet().time, doc="Ra Number of Free Convection")
        def Ra_number_eqn(b, t):
            return (
                b.N_Ra_root6[t]
                == b.const_Ra_root6
                * sqrt(b.drum_do + 2 * b.insulation_thickness)
                * (b.temp_insulation_outside[t] - b.temperature_ambient[t]) ** 0.166667
            )

        # Nu number equation
        @self.Constraint(self.flowsheet().time, doc="Nu Number of Free Convection")
        def Nu_number_eqn(b, t):
            return b.N_Nu[t] == (0.6 + b.const_Nu * b.N_Ra_root6[t]) ** 2

        # Free convection coefficient based on the drum metal outside diameter
        @self.Constraint(
            self.flowsheet().time,
            doc="Free Convection Heat Transfer Coefficient"
            "between Insulation Wall and Ambient",
        )
        def h_free_conv_eqn(b, t):
            return (
                b.heat_transfer_free_conv[t] == b.N_Nu[t] * b.therm_cond_air / b.drum_do
            )

        @self.Constraint(self.flowsheet().time, doc="Heat Loss of Water")
        def heat_loss_eqn(b, t):
            return b.heat_duty[t] == b.drum_area * b.heat_transfer_in[t] * (
                b.drum_wall_temperature[t, b.radial_domain.first()]
                - b.control_volume.properties_out[t].temperature
            )

        # Calculate mechanical and thermal stresses based on
        # EN 13445 Standard
        # Integer indexing for radius domain
        self.rindex = Param(
            self.radial_domain,
            initialize=1,
            mutable=True,
            doc="Integer Indexing for Radius Domain",
        )

        # calculate integral point for mean temperature in the wall
        @self.Expression(self.flowsheet().time, doc="Mean Temperature across the Wall")
        def mean_temperature(b, t):
            return (
                2
                * (b.radial_domain.at(2) - b.radial_domain.at(1))
                / (b.drum_ro**2 - b.drum_ri**2)
                * (
                    sum(
                        0.5
                        * (
                            b.radial_domain.at(i - 1)
                            * b.drum_wall_temperature[t, b.radial_domain.at(i - 1)]
                            + b.radial_domain.at(i)
                            * b.drum_wall_temperature[t, b.radial_domain.at(i)]
                        )
                        for i in range(2, len(b.radial_domain) + 1)
                    )
                )
            )

        for index_r, value_r in enumerate(self.radial_domain, 1):
            self.rindex[value_r] = index_r

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Discrete Point Mean Temperature",
        )
        def discrete_mean_temperature(b, t, r):
            if b.rindex[r].value == 1:
                return b.drum_wall_temperature[t, b.radial_domain.first()]
            else:
                return (
                    2
                    * (b.radial_domain.at(2) - b.radial_domain.at(1))
                    / (b.radial_domain.at(b.rindex[r].value) ** 2 - b.drum_ri**2)
                    * (
                        sum(
                            0.5
                            * (
                                b.radial_domain.at(j - 1)
                                * b.drum_wall_temperature[t, b.radial_domain.at(j - 1)]
                                + b.radial_domain.at(j)
                                * b.drum_wall_temperature[t, b.radial_domain.at(j)]
                            )
                            for j in range(2, b.rindex[r].value + 1)
                        )
                    )
                )

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Thermal Stress at Radial Direction for Drum",
        )
        def therm_sigma_r(b, t, r):
            if r == b.radial_domain.first() or r == b.radial_domain.last():
                return 0
            else:
                return (
                    0.5
                    * b.Young_modulus
                    * b.coefficient_therm_expansion
                    / (1 - b.Poisson_ratio)
                    * (
                        (1 - b.drum_ri**2 / r**2)
                        * (b.mean_temperature[t] - b.discrete_mean_temperature[t, r])
                    )
                )

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Thermal Stress at Circumferential Direction" "for Drum",
        )
        def therm_sigma_theta(b, t, r):
            return (
                0.5
                * b.Young_modulus
                * b.coefficient_therm_expansion
                / (1 - b.Poisson_ratio)
                * (
                    (1 + b.drum_ri**2 / r**2) * b.mean_temperature[t]
                    + (1 - b.drum_ri**2 / r**2) * b.discrete_mean_temperature[t, r]
                    - 2 * b.drum_wall_temperature[t, r]
                )
            )

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Thermal Stress at Axial Direction for Drum",
        )
        def therm_sigma_z(b, t, r):
            return (
                b.Young_modulus
                * b.coefficient_therm_expansion
                / (1 - b.Poisson_ratio)
                * (b.mean_temperature[t] - b.drum_wall_temperature[t, r])
            )

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Mechanical Stress at Radial Direction for Drum",
        )
        def mech_sigma_r(b, t, r):
            if r == b.radial_domain.first():
                return 1e-6 * (-b.control_volume.properties_out[t].pressure)
            elif r == b.radial_domain.last():
                return 1e-6 * (-b.pressure_amb)
            else:
                return 0.1 * (
                    1e-5
                    * (
                        b.control_volume.properties_out[t].pressure * b.drum_ri**2
                        - b.pressure_amb * b.drum_ro**2
                    )
                    / (b.drum_ro**2 - b.drum_ri**2)
                    + (
                        1e-5
                        * (b.pressure_amb - b.control_volume.properties_out[t].pressure)
                        * b.drum_ri**2
                        * b.drum_ro**2
                        / (r**2 * (b.drum_ro**2 - b.drum_ri**2))
                    )
                )

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Mechanical Stress at Circumferential Direction" "for Drum",
        )
        def mech_sigma_theta(b, t, r):
            return 0.1 * (
                1e-5
                * (
                    b.control_volume.properties_out[t].pressure * b.drum_ri**2
                    - b.pressure_amb * b.drum_ro**2
                )
                / (b.drum_ro**2 - b.drum_ri**2)
                - (
                    1e-5
                    * (b.pressure_amb - b.control_volume.properties_out[t].pressure)
                    * b.drum_ri**2
                    * b.drum_ro**2
                    / (r**2 * (b.drum_ro**2 - b.drum_ri**2))
                )
            )

        @self.Expression(
            self.flowsheet().time, doc="Mechanical Stress at Axial Direction" "for Drum"
        )
        def mech_sigma_z(b, t):
            return 0.1 * (
                1e-5
                * (
                    b.control_volume.properties_out[t].pressure * b.drum_ri**2
                    - b.pressure_amb * b.drum_ro**2
                )
                / (b.drum_ro**2 - b.drum_ri**2)
            )

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Principal Structural Stress" "at Radial Direction for Drum",
        )
        def sigma_r(b, t, r):
            if r == b.radial_domain.first():
                return 1e-6 * (-b.control_volume.properties_out[t].pressure)
            elif r == b.radial_domain.last():
                return 1e-6 * (-b.pressure_amb)
            else:
                return b.mech_sigma_r[t, r] + b.therm_sigma_r[t, r]

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Principal Structural Stress" "at Circumferential Direction for Drum",
        )
        def sigma_theta(b, t, r):
            return b.mech_sigma_theta[t, r] + b.therm_sigma_theta[t, r]

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Principal Structural Stress" "at Axial Direction for Drum",
        )
        def sigma_z(b, t, r):
            return b.mech_sigma_z[t] + b.therm_sigma_z[t, r]

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Equivalent von Mises Stress for Drum",
        )
        def sigma_von_Mises(b, t, r):
            return sqrt(
                b.sigma_r[t, r] ** 2
                + b.sigma_theta[t, r] ** 2
                + b.sigma_z[t, r] ** 2
                - (
                    b.sigma_r[t, r] * b.sigma_theta[t, r]
                    + b.sigma_r[t, r] * b.sigma_z[t, r]
                    + b.sigma_theta[t, r] * b.sigma_z[t, r]
                )
            )

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Variation Principal Stress"
            "between Radial-Circumferential Directions for Drum",
        )
        def delta_sigma_r_theta(b, t, r):
            return abs(b.sigma_r[t, r] - b.sigma_theta[t, r])

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Variation Principal Stress"
            "between Circumferential-Axial Directions for Drum",
        )
        def delta_sigma_theta_z(b, t, r):
            return abs(b.sigma_theta[t, r] - b.sigma_z[t, r])

        @self.Expression(
            self.flowsheet().time,
            self.radial_domain,
            doc="Variation Principal Stress" "between Axial-Radial Directions for Drum",
        )
        def delta_sigma_z_r(b, t, r):
            return abs(b.sigma_z[t, r] - b.sigma_r[t, r])

        # Calculate stress based on EN12952 standard
        # -----------------------------------------------------------------
        # mechanical stress of circumferential direction
        # mean radius
        r_ms_drum = self.drum_ri + self.drum_thickness / 2

        # thickness of downcomer //m
        self.downcomer_thickness = Param(
            initialize=0.0254, mutable=True, doc="Thickness of Downcomer"
        )

        # mean diameter of downcomer
        # pipe_d = pipe_d(inner) + self.downcomer_thickness ==
        # pipe_d(outer)-self.downcomer_thickness
        pipe_d = self.downcomer_diameter + self.downcomer_thickness

        # mechanical coefficients
        k_m_A = (
            -1.14 * (self.downcomer_thickness / self.drum_thickness) ** 2
            - 0.89 * (self.downcomer_thickness / self.drum_thickness)
            + 1.43
        )
        k_m_B = (
            0.326 * (self.downcomer_thickness / self.drum_thickness) ** 2
            - 0.59 * (self.downcomer_thickness / self.drum_thickness)
            + 1.08
        )
        k_m_C = (
            pipe_d / (2 * r_ms_drum) * sqrt((2 * r_ms_drum) / (2 * self.drum_thickness))
        )
        k_m = 2.2 + exp(k_m_A) * k_m_C**k_m_B

        # thermal stress concentration factor
        k_t_A = pipe_d / (2 * r_ms_drum)
        # heat transfer coefficient:3000 for water , 1000 for steam
        k_t_B = 1000
        k_t = sqrt(
            (
                2
                - (k_t_B + 2700) / (k_t_B + 1700) * k_t_A
                + k_t_B / (k_t_B + 1700) * (exp(-7 * k_t_A) - 1)
            )
            ** 2
            + 0.81 * k_t_A**2
        )

        # mechanical stress at circumferential direction
        @self.Expression(
            self.flowsheet().time,
            doc="Mechanical Stress "
            "at Circumferential Direction"
            "for Drum (EN 12952-3)",
        )
        def sigma_p(b, t):
            return (
                0.1
                * 1e-5
                * b.control_volume.properties_out[t].pressure
                * r_ms_drum
                / self.drum_thickness
            )

        # thermal stress at circumferential direction
        @self.Expression(
            self.flowsheet().time,
            doc="Thermal Stress at Circumferential" "Direction for Drum (EN 12952-3)",
        )
        def sigma_t(b, t):
            delta_T = (
                b.mean_temperature[t]
                - b.drum_wall_temperature[t, b.radial_domain.first()]
            )
            return (
                b.coefficient_therm_expansion
                * b.Young_modulus
                / (1 - b.Poisson_ratio)
                * delta_T
            )

        # calculate stress at 2 locations at the hole
        # stress at crotch corner P1 and location P2

        # mechanical stress by pressure at crotch corner
        @self.Expression(
            self.flowsheet().time, doc="Mechanical Stress at Crotch Corner" "for Drum"
        )
        def sigma_p_P1(b, t):
            return b.sigma_p[t] * k_m

        # mechanical stress at location P2
        @self.Expression(
            self.flowsheet().time,
            doc="Mechanical Stress at" "Critical Point P2 for Drum",
        )
        def sigma_p_P2(b, t):
            return b.sigma_p[t] * k_m / 5

        # thermal stress at crotch corner
        @self.Expression(
            self.flowsheet().time, doc="Thermal Stress at Crotch Corner for Drum"
        )
        def sigma_t_P1(b, t):
            return b.sigma_t[t] * k_t

        # thermal stress at location P2
        @self.Expression(
            self.flowsheet().time, doc="Thermal Stress" "at Critical Point P2 for Drum"
        )
        def sigma_t_P2(b, t):
            return b.sigma_t[t] * k_t

        # total circumferential stress with notch effect
        # crotch corner P1
        @self.Expression(
            self.flowsheet().time,
            doc="Circumferential Stress" "at Crotch Corner for Drum",
        )
        def sigma_theta_P1(b, t):
            return b.sigma_p_P1[t] + b.sigma_t_P1[t]

        # location P2
        @self.Expression(
            self.flowsheet().time,
            doc="Circumferential Stress" "at Critical Point P2 for Drum",
        )
        def sigma_theta_P2(b, t):
            return b.sigma_p_P2[t] + b.sigma_t_P2[t]

        # total stress with notch effect // f1 - f2
        # crotch corner
        @self.Expression(
            self.flowsheet().time, doc="Total Stress at Crotch Corner for Drum"
        )
        def sigma_notch_P1(b, t):
            return (
                b.sigma_theta_P1[t] + 1e-6 * b.control_volume.properties_out[t].pressure
            )

        # location P2
        @self.Expression(
            self.flowsheet().time, doc="Total Stress at Critial Point P2 for Header"
        )
        def sigma_notch_P2(b, t):
            return (
                b.sigma_theta_P2[t] + 1e-6 * b.control_volume.properties_out[t].pressure
            )

        # Von Mises equivalent stress
        # VM stress at crotch corner
        @self.Expression(
            self.flowsheet().time,
            doc="Equivalent von Mises Stress" "at Crotch Corner for Drum",
        )
        def sigma_eff_P1(b, t):
            sigma_comer_P1 = b.sigma_theta_P1[t]
            p_in = b.control_volume.properties_out[t].pressure
            return sqrt(
                sigma_comer_P1**2
                + (-1e-6 * p_in) ** 2
                + (-1e-6 * p_in) ** 2
                - (
                    (-1e-6 * p_in) * sigma_comer_P1
                    + (-1e-6 * p_in) * sigma_comer_P1
                    + (-1e-6 * p_in) * (-1e-6 * p_in)
                )
            )

        # VM stress at location P2
        @self.Expression(
            self.flowsheet().time,
            doc="Equivalent von Mises Stress" "at Critical Point P2 for Drum",
        )
        def sigma_eff_P2(b, t):
            sigma_comer_P2 = b.sigma_theta_P2[t]
            p_in = b.control_volume.properties_out[t].pressure
            return sqrt(
                sigma_comer_P2**2
                + (-1e-6 * p_in) ** 2
                + (-1e-6 * p_in) ** 2
                - (
                    (-1e-6 * p_in) * sigma_comer_P2
                    + (-1e-6 * p_in) * sigma_comer_P2
                    + (-1e-6 * p_in) * (-1e-6 * p_in)
                )
            )

    def set_initial_condition(self):
        if self.config.dynamic is True:
            self.control_volume.material_accumulation[:, :, :].value = 0
            self.control_volume.energy_accumulation[:, :].value = 0
            self.dTdt[:, :].value = 0
            self.control_volume.material_accumulation[0, :, :].fix(0)
            self.control_volume.energy_accumulation[0, :].fix(0)
            self.dTdt[0, :].fix(0)

    def initialize_build(
        blk,
        state_args_feedwater=None,
        state_args_water_steam=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Drum initialization routine.
        Keyword Arguments:
        state_args_feedwater : a dict of arguments to be passed to the property
        package(s) for the control_volume of the model to
        provide an initial state for initialization
        (see documentation of the specific property package)
        (default = None).

        state_args_steam : a dict of arguments to be passed to the property
        package(s) for the control_volume of the model to
        provide an initial state for initialization
        (see documentation of the specific property package)
        (default = None).

        outlvl : sets output level of initialisation routine
                 * 0 = no output (default)
                 * 1 = return solver state for each step in routine
                 * 2 = return solver state for each step in subroutines
                 * 3 = include solver output information (tee=True)

        optarg : solver options dictionary object (default=None, use
                 default solver options)

        solver : str indicating which solver to use during
                 initialization (default = None, use default solver)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        init_log.info_low("Starting Initialization...")
        # fix FeedWater Inlet
        flags_fw = fix_state_vars(blk.mixer.FeedWater_state, state_args_feedwater)

        blk.flash.initialize(
            state_args_water_steam=state_args_water_steam,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
        )
        init_log.info("Initialization Step 1 Complete. Flash Model Solved")

        # Copy the state values from flash model liquid outlet to
        # mixer model saturated water inlet
        blk.mixer.SaturatedWater.flow_mol[:].value = blk.flash.liq_outlet.flow_mol[
            0
        ].value
        blk.mixer.SaturatedWater.pressure[:].value = blk.flash.liq_outlet.pressure[
            0
        ].value
        blk.mixer.SaturatedWater.enth_mol[:].value = blk.flash.liq_outlet.enth_mol[
            0
        ].value

        # Since the mixer model requires the same inlet pressure,
        # unfix feed water pressure
        # use the water_steam inlet pressure as mixer outlet pressure
        blk.mixer.FeedWater.pressure.unfix()

        blk.mixer.initialize(outlvl=outlvl, optarg=optarg, solver=solver)
        init_log.info("Initialization Step 2 Complete. Mixer Model Solved")

        # Copy the state values from mixer model outlet to control volume inlet
        blk.control_volume.properties_in[:].flow_mol.value = blk.mixer.outlet.flow_mol[
            0
        ].value
        blk.control_volume.properties_in[:].enth_mol.value = blk.mixer.outlet.enth_mol[
            0
        ].value
        blk.control_volume.properties_in[:].pressure.value = blk.mixer.outlet.pressure[
            0
        ].value

        blk.control_volume.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, hold_state=False
        )
        init_log.info("Initialization Step 3 Complete. Control Volume Initialized")

        # fix flash model inlet
        flags_steam = fix_state_vars(blk.flash.mixed_state, state_args_water_steam)

        # set initial values for T
        r_mid = value((blk.radial_domain.first() + blk.radial_domain.last()) / 2)
        # assume outside wall temperature is 1 K lower than fluid temperature
        T_out = value(blk.control_volume.properties_in[0].temperature - 1)
        T_mid = value((T_out + blk.control_volume.properties_in[0].temperature) / 2)
        slope = value(
            (T_out - blk.control_volume.properties_in[0].temperature)
            / (blk.radial_domain.last() - blk.radial_domain.first())
            / 3
        )
        for x in blk.radial_domain:
            blk.drum_wall_temperature[:, x].fix(T_mid + slope * (x - r_mid))
        blk.drum_wall_temperature[:, :].unfix()

        # Fix outlet enthalpy and pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].pressure.fix(
                value(blk.control_volume.properties_in[0].pressure) - 5000.0
            )
            blk.control_volume.properties_out[t].enth_mol.fix(
                value(blk.control_volume.properties_in[0].enth_mol)
            )
        blk.pressure_change_total_eqn.deactivate()
        blk.heat_loss_eqn.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high(
            "Initialization Step 4 Complete {}.".format(idaeslog.condition(res))
        )

        # Unfix outlet enthalpy and pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].pressure.unfix()
            blk.control_volume.properties_out[t].enth_mol.unfix()
        blk.pressure_change_total_eqn.activate()
        blk.heat_loss_eqn.activate()

        # Finally solve model with all constraints activated
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 5 {}.".format(idaeslog.condition(res)))
        revert_state_vars(blk.mixer.FeedWater_state, flags_fw)
        revert_state_vars(blk.flash.mixed_state, flags_steam)
        init_log.info("Initialization Complete.")

    def calculate_scaling_factors(self):
        for v in self.deltaP_gravity.values():
            if iscale.get_scaling_factor(v, warning=True) is None:
                iscale.set_scaling_factor(v, 1e-3)

        for v in self.deltaP_contraction.values():
            if iscale.get_scaling_factor(v, warning=True) is None:
                iscale.set_scaling_factor(v, 1e-3)

        for t, c in self.pressure_change_contraction_eqn.items():
            sf = iscale.get_scaling_factor(
                self.deltaP_contraction[t], default=1, warning=True
            )
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        for t, c in self.pressure_change_gravity_eqn.items():
            sf = iscale.get_scaling_factor(
                self.deltaP_gravity[t], default=1, warning=True
            )
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        for t, c in self.pressure_change_total_eqn.items():
            sf = iscale.get_scaling_factor(self.deltaP[t], default=1, warning=True)
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        for t, c in self.connection_material_balance.items():
            sf = iscale.get_scaling_factor(1e-4, default=1)
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        for t, c in self.connection_enthalpy_balance.items():
            sf = iscale.get_scaling_factor(1e-4, default=1)
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        for t, c in self.connection_pressure_balance.items():
            sf = iscale.get_scaling_factor(1e-6, default=1)
            iscale.constraint_scaling_transform(c, sf, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Metrics object
"""
import pytest
import pandas as pd

from idaes.core.surrogate.metrics import compute_fit_metrics
from idaes.core.surrogate import AlamoSurrogate

# For this test we will use a simple z = x function and calculate metrics
# Measured (test) data will include a fixed offset (i.e. z = x + Err)
# Data will be collected symmetrically around 0, so that mean z1 == 0
ERR = 0.1  # offset to apply to check metrics
Np = 50  # number of points above 0 to use in check
N = 2 * Np + 1  # total number of points to use in check


@pytest.fixture
def metrics():
    # Create dataset
    x = []
    z = []
    for i in range(N):
        v = 0.1 * i - Np / 10
        x.append(v)
        z.append(v + ERR)  # add offset of 0.1
    dataset = pd.DataFrame({"x1": x, "z1": z})

    # Create a dummy ALAMO surrogate to use for testing
    alm_obj = AlamoSurrogate(
        surrogate_expressions={"z1": "z1 == x1"},
        input_labels=["x1"],
        output_labels=["z1"],
    )

    metrics = compute_fit_metrics(surrogate=alm_obj, dataframe=dataset)

    return metrics


@pytest.mark.unit
def test_compute_metrics(metrics):

    # All data should have an offset of ERR, so error is known
    assert metrics["z1"]["SSE"] == pytest.approx(ERR**2 * N, rel=1e-12)
    assert metrics["z1"]["MSE"] == pytest.approx(ERR**2, rel=1e-12)
    assert metrics["z1"]["RMSE"] == pytest.approx(ERR, rel=1e-12)

    # Max and mean errors should be equal to ERR
    assert metrics["z1"]["MAE"] == pytest.approx(ERR, rel=1e-12)
    assert metrics["z1"]["maxAE"] == pytest.approx(ERR, rel=1e-12)

    # Calculate SST
    # Average measured z1 is equal to ERR, thus measured z1-z1_mean = x1
    sst = 0
    for i in range(int(Np)):
        sst += 2 * (Np / 10 - 0.1 * i) ** 2

    assert metrics["z1"]["R2"] == pytest.approx(1 - ERR**2 * N / sst, rel=1e-12)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Pipe model for water or steam, Liq or Vap phase must be provided (mixed phase
not supported).

main equations:

* Heat is given (zero if adiabatic)
* Calculate pressure change due to friction and gravity
* Assume enthalpy_in = enthalpy_out + heat

Created: April 2019 by Jinliang Ma
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
from pyomo.environ import value, Var, Reference

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)

from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.constants import Constants as const
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog

__author__ = "Boiler Subsystem Team (J. Ma, M. Zamarripa)"
__version__ = "2.0.0"


@declare_process_block_class("WaterPipe")
class WaterPipeData(UnitModelBlockData):
    """
    Water or steam pipe Unit Class
    """

    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.componentPhase,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of material balance should be constructed,
**default** - MaterialBalanceType.componentPhase.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.enthalpyTotal,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single ethalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - ethalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_heat_transfer",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Heat transfer term construction flag",
            doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "contraction_expansion_at_end",
        ConfigValue(
            default="None",
            domain=In(["None", "contraction", "expansion"]),
            description="Any contraction or expansion at the end",
            doc="Define if pressure drop due to contraction"
            " or expansion needs to be considered",
        ),
    )
    CONFIG.declare(
        "water_phase",
        ConfigValue(
            default="Liq",
            domain=In(["Liq", "Vap"]),
            description="Water phase",
            doc="""Define water phase for property calls,
mixed phase not supported""",
        ),
    )

    def build(self):
        """
        Begin building model
        """
        # Call UnitModel.build to setup dynamics
        super(WaterPipeData, self).build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_geometry()

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type,
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer,
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type, has_pressure_change=True
        )

        # Add Ports
        self.add_inlet_port()
        self.add_outlet_port()

        # Add object references
        self.volume = Reference(self.control_volume.volume)

        # Set references to balance terms at unit level
        if (
            self.config.has_heat_transfer is True
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            self.heat_duty = Reference(self.control_volume.heat)

        if (
            self.config.has_pressure_change is True
            and self.config.momentum_balance_type != "none"
        ):
            self.deltaP = Reference(self.control_volume.deltaP)

        # Set Unit Geometry and Volume
        self._set_geometry()

        # Construct performance equations
        self._make_performance()

    def _set_geometry(self):
        """
        Define the geometry of the unit as necessary
        """
        # Number of pipe
        self.number_of_pipes = Var(initialize=4, doc="Number of pipes")
        # Length of pipe
        self.length = Var(initialize=10.0, doc="Total length of the straight pipe")
        # Elevation change of pipe, elevation of outlet - elevation of inlet
        self.elevation_change = Var(
            initialize=10.0, doc="Change of elevation from inlet to outlet"
        )
        # Inside diameter of pipe
        self.diameter = Var(initialize=0.6, doc="Inside diameter of pipe")

        if not self.config.contraction_expansion_at_end == "None":
            self.area_ratio = Var(
                initialize=1, doc="Cross section area ratio of exit end to pipe"
            )

        # Volume constraint
        @self.Constraint(self.flowsheet().time, doc="Total volume of all pipes")
        def volume_eqn(b, t):
            return (
                b.volume[t]
                == 0.25 * const.pi * b.diameter**2 * b.length * b.number_of_pipes
            )

    def _make_performance(self):
        """
        Define constraints which describe the behaviour of the unit model.
        """
        phase = self.config.water_phase

        # Add performance variables
        # Velocity of fluid inside pipe
        self.velocity = Var(
            self.flowsheet().time, initialize=10.0, doc="Fluid velocity inside pipe"
        )

        # Reynolds number
        self.N_Re = Var(
            self.flowsheet().time, initialize=10000.0, doc="Reynolds number"
        )

        # Darcy friction factor
        self.friction_factor_darcy = Var(
            self.flowsheet().time, initialize=0.005, doc="Darcy friction factor"
        )

        # Correction factor for pressure drop due to friction
        self.fcorrection_dp = Var(
            initialize=1.0, doc="Correction factor for pressure drop"
        )

        # Pressure change due to friction
        self.deltaP_friction = Var(
            self.flowsheet().time,
            initialize=-1.0,
            doc="Pressure change due to friction",
        )

        # Pressure change due to gravity
        self.deltaP_gravity = Var(
            self.flowsheet().time, initialize=0.0, doc="Pressure change due to gravity"
        )

        # Pressure change due to area change at end
        self.deltaP_area_change = Var(
            self.flowsheet().time,
            initialize=0.0,
            doc="Pressure change due to area change",
        )

        # Equation for calculating velocity
        @self.Constraint(self.flowsheet().time, doc="Velocity of fluid inside pipe")
        def velocity_eqn(b, t):
            return (
                b.velocity[t] * 0.25 * const.pi * b.diameter**2 * b.number_of_pipes
                == b.control_volume.properties_in[t].flow_vol
            )

        # Equation for calculating Reynolds number
        @self.Constraint(self.flowsheet().time, doc="Reynolds number")
        def Reynolds_number_eqn(b, t):
            return (
                b.N_Re[t] * b.control_volume.properties_in[t].visc_d_phase[phase]
                == b.diameter
                * b.velocity[t]
                * b.control_volume.properties_in[t].dens_mass_phase[phase]
            )

        # Friction factor expression depending on laminar or turbulent flow
        @self.Constraint(
            self.flowsheet().time,
            doc="Darcy friction factor as" " a function of Reynolds number",
        )
        def friction_factor_darcy_eqn(b, t):
            return (
                b.friction_factor_darcy[t] * b.N_Re[t] ** (0.25)
                == 0.3164 * b.fcorrection_dp
            )

        # Pressure change equation for friction
        @self.Constraint(self.flowsheet().time, doc="Pressure change due to friction")
        def pressure_change_friction_eqn(b, t):
            return (
                b.deltaP_friction[t] * b.diameter
                == -0.5
                * b.control_volume.properties_in[t].dens_mass_phase[phase]
                * b.velocity[t] ** 2
                * b.friction_factor_darcy[t]
                * b.length
            )

        # Pressure change equation for gravity
        @self.Constraint(self.flowsheet().time, doc="Pressure change due to gravity")
        def pressure_change_gravity_eqn(b, t):
            return (
                b.deltaP_gravity[t]
                == -const.acceleration_gravity
                * b.control_volume.properties_in[t].dens_mass_phase[phase]
                * b.elevation_change
            )

        # Pressure change equation for contraction or expansion
        @self.Constraint(self.flowsheet().time, doc="Pressure change due to gravity")
        def pressure_change_area_change_eqn(b, t):
            if self.config.contraction_expansion_at_end == "contraction":
                return (
                    b.deltaP_area_change[t]
                    == -(0.1602 * b.area_ratio**2 - 0.646 * b.area_ratio + 1.4858)
                    * 0.5
                    * b.control_volume.properties_out[t].dens_mass_phase[phase]
                    * (b.velocity[t] / b.area_ratio) ** 2
                    + 0.5
                    * b.control_volume.properties_out[t].dens_mass_phase[phase]
                    * b.velocity[t] ** 2
                )
            elif self.config.contraction_expansion_at_end == "expansion":
                return (
                    b.deltaP_area_change[t]
                    == b.control_volume.properties_out[t].dens_mass_phase[phase]
                    * b.velocity[t] ** 2
                    * (b.area_ratio - 1)
                    / b.area_ratio**2
                )
            else:
                return b.deltaP_area_change[t] == 0

        # Total pressure change equation
        @self.Constraint(self.flowsheet().time, doc="Pressure drop")
        def pressure_change_total_eqn(b, t):
            return b.deltaP[t] == (
                b.deltaP_friction[t] + b.deltaP_gravity[t] + b.deltaP_area_change[t]
            )

    def set_initial_condition(self):
        if self.config.dynamic is True:
            self.control_volume.material_accumulation[:, :, :].value = 0
            self.control_volume.energy_accumulation[:, :].value = 0
            self.control_volume.material_accumulation[0, :, :].fix(0)
            self.control_volume.energy_accumulation[0, :].fix(0)

    def initialize_build(
        blk, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        WaterPipe initialization routine.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) for the control_volume of the model to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            outlvl : sets output level of initialisation routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        flags = blk.control_volume.initialize(
            outlvl=outlvl + 1, optarg=optarg, solver=solver, state_args=state_args
        )
        init_log.info_high("Initialization Step 1 Complete.")
        # Fix outlet enthalpy and pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].pressure.fix(
                value(blk.control_volume.properties_in[t].pressure)
            )
        blk.pressure_change_total_eqn.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # Unfix outlet enthalpy and pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].pressure.unfix()
        blk.pressure_change_total_eqn.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))
        blk.control_volume.release_state(flags, outlvl)
        init_log.info("Initialization Complete.")

    def calculate_scaling_factors(self):
        pass



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring


def compute_fit_metrics(surrogate, dataframe):
    """
    Compute the fit metrics for the surrogate against the
    data in dataframe. The surrogate will be evaluated
    for the input values in the dataframe, and the results
    will be compared against the output values in the dataframe
    to provide the metrics.

    Args:
       surrogate : surrogate object (derived from SurrogateBase)
          This is the surrogate object we want to evaluate for the comparison
       dataframe : pandas DataFrame
          The dataframe that contains the inputs and outputs we want to use
          in the evaluation.

    Returns:
        dict-of-dicts with outer keys representing output labels and inner keys
        representing metrics for that output.
    """
    y = dataframe[surrogate.output_labels()]
    f = surrogate.evaluate_surrogate(dataframe)
    assert f.columns.to_list() == surrogate.output_labels()

    y_mean = y.mean(axis=0)
    SST = ((y - y_mean) ** 2).sum(axis=0)
    SSE = ((y - f) ** 2).sum(axis=0)

    R2 = 1 - SSE / SST
    MAE = (y - f).abs().mean(axis=0)
    maxAE = (y - f).abs().max(axis=0)
    MSE = ((y - f) ** 2).mean(axis=0)
    RMSE = MSE**0.5

    # Reorder indices to have output first
    metrics = {}
    for o in surrogate.output_labels():
        metrics[o] = {
            "RMSE": RMSE[o],
            "MSE": MSE[o],
            "MAE": MAE[o],
            "maxAE": maxAE[o],
            "SSE": SSE[o],
            "R2": R2[o],
        }

    return metrics



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
General tests for generic properties with Henry components present

Author: Andrew Lee
"""
# Import Python libraries
import pytest

# Import Pyomo components
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Set,
    value,
    units as pyunits,
)

# Import IDAES cores
from idaes.core import LiquidPhase, VaporPhase, Component
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.core.solvers import get_solver

from idaes.models.properties.modular_properties.base.generic_property import (
    _init_Pbub,
    _init_Pdew,
    _init_Tbub,
    _init_Tdew,
)
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.forms import fugacity
from idaes.models.properties.modular_properties.phase_equil.henry import ConstantH
import idaes.models.properties.modular_properties.pure.Perrys as Perrys
import idaes.models.properties.modular_properties.pure.RPP4 as RPP4
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    IdealBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.henry import HenryType

import idaes.logger as idaeslog

# Set up logger
_log = idaeslog.getLogger(__name__)
solver = get_solver("ipopt_v2")


class TestNoHenryComps(object):
    configuration = {
        # Specifying components
        "components": {
            "A": {
                "type": Component,
                "enth_mol_liq_comp": Perrys,
                "enth_mol_ig_comp": RPP4,
                "pressure_sat_comp": RPP4,
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
                "parameter_data": {
                    "mw": 78.1136e-3,
                    "pressure_crit": 48.9e5,
                    "temperature_crit": 562.2,
                    "cp_mol_ig_comp_coeff": {
                        "A": -3.392e1,
                        "B": 4.739e-1,
                        "C": -3.017e-4,
                        "D": 7.130e-8,
                    },
                    "cp_mol_liq_comp_coeff": {
                        "1": 1.29e2,
                        "2": -1.7e-1,
                        "3": 6.48e-4,
                        "4": 0,
                        "5": 0,
                    },
                    "enth_mol_form_liq_comp_ref": 49.0e3,
                    "enth_mol_form_vap_comp_ref": 82.9e3,
                    "pressure_sat_comp_coeff": {
                        "A": -6.98273,
                        "B": 1.33213,
                        "C": -2.62863,
                        "D": -3.33399,
                    },
                },
            },
            "B": {
                "type": Component,
                "enth_mol_liq_comp": Perrys,
                "enth_mol_ig_comp": RPP4,
                "pressure_sat_comp": RPP4,
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
                "parameter_data": {
                    "mw": 92.1405e-3,
                    "pressure_crit": 41e5,
                    "temperature_crit": 591.8,
                    "cp_mol_ig_comp_coeff": {
                        "A": -2.435e1,
                        "B": 5.125e-1,
                        "C": -2.765e-4,
                        "D": 4.911e-8,
                    },
                    "cp_mol_liq_comp_coeff": {
                        "1": 1.40e2,
                        "2": -1.52e-1,
                        "3": 6.95e-4,
                        "4": 0,
                        "5": 0,
                    },
                    "enth_mol_form_liq_comp_ref": 12.0e3,
                    "enth_mol_form_vap_comp_ref": 50.1e3,
                    "pressure_sat_comp_coeff": {
                        "A": -7.28607,
                        "B": 1.38091,
                        "C": -2.83433,
                        "D": -2.79168,
                    },
                },
            },
        },
        # Specifying phases
        "phases": {
            "Liq": {"type": LiquidPhase, "equation_of_state": Ideal},
            "Vap": {"type": VaporPhase, "equation_of_state": Ideal},
        },
        # Declare a base units dict to save code later
        "base_units": {
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        # Specifying state definition
        "state_definition": FTPx,
        "state_bounds": {
            "flow_mol": (0, 100, 1000),
            "temperature": (273.15, 300, 450),
            "pressure": (5e4, 1e5, 1e6),
        },
        "pressure_ref": 1e5,
        "temperature_ref": 300,
        # Defining phase equilibria
        "phases_in_equilibrium": [("Vap", "Liq")],
        "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
        "bubble_dew_method": IdealBubbleDew,
    }

    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**self.configuration)

        model.props = model.params.build_state_block([1], defined_state=True)

        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(368)
        model.props[1].pressure.fix(101325)
        model.props[1].mole_frac_comp["A"].fix(0.5)
        model.props[1].mole_frac_comp["B"].fix(0.5)

        # Trigger construction of some things we will test later
        model.props[1].pressure_bubble
        model.props[1].pressure_dew

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 2
        for i in model.params.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.params.Liq.is_liquid_phase()
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 2
        for i in model.params.component_list:
            assert i in ["A", "B"]
            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 4
        for i in model.params._phase_component_set:
            assert i in [("Liq", "A"), ("Liq", "B"), ("Vap", "A"), ("Vap", "B")]

        assert model.params.config.state_definition == FTPx

        assert model.params.config.state_bounds == {
            "flow_mol": (0, 100, 1000),
            "temperature": (273.15, 300, 450),
            "pressure": (5e4, 1e5, 1e6),
        }

        assert model.params.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE
        }

        assert isinstance(model.params.phase_equilibrium_idx, Set)
        assert len(model.params.phase_equilibrium_idx) == 2
        for i in model.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2"]

        assert model.params.phase_equilibrium_list == {
            "PE1": ["A", ("Vap", "Liq")],
            "PE2": ["B", ("Vap", "Liq")],
        }

        assert model.params.pressure_ref.value == 1e5
        assert model.params.temperature_ref.value == 300

    @pytest.mark.unit
    def test_init_bubble_temperature(self, model):
        _init_Tbub(model.props[1], pyunits.K)

        assert pytest.approx(365.35, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.7137, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2863, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")]
        )

    @pytest.mark.unit
    def test_init_dew_temperature(self, model):
        _init_Tdew(model.props[1], pyunits.K)

        assert pytest.approx(372.02, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2909, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.7091, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")]
        )

    @pytest.mark.unit
    def test_init_bubble_pressure(self, model):
        _init_Pbub(model.props[1])

        assert pytest.approx(109479, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")]
        )

    @pytest.mark.unit
    def test_init_dew_pressure(self, model):
        _init_Pdew(model.props[1])

        assert pytest.approx(89820, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")]
        )

    @pytest.mark.component
    def test_solve_vle(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

        assert pytest.approx(365.35, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.7137, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2863, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")]
        )

        assert pytest.approx(372.02, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2909, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.7091, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")]
        )

        assert pytest.approx(109479, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")]
        )

        assert pytest.approx(89820, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")]
        )


configuration = {
    # Specifying components
    "components": {
        "A": {
            "type": Component,
            "enth_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            "parameter_data": {
                "mw": 78.1136e-3,
                "pressure_crit": 48.9e5,
                "temperature_crit": 562.2,
                "cp_mol_ig_comp_coeff": {
                    "A": -3.392e1,
                    "B": 4.739e-1,
                    "C": -3.017e-4,
                    "D": 7.130e-8,
                },
                "cp_mol_liq_comp_coeff": {
                    "1": 1.29e2,
                    "2": -1.7e-1,
                    "3": 6.48e-4,
                    "4": 0,
                    "5": 0,
                },
                "enth_mol_form_liq_comp_ref": 49.0e3,
                "enth_mol_form_vap_comp_ref": 82.9e3,
                "pressure_sat_comp_coeff": {
                    "A": -6.98273,
                    "B": 1.33213,
                    "C": -2.62863,
                    "D": -3.33399,
                },
            },
        },
        "B": {
            "type": Component,
            "enth_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            "parameter_data": {
                "mw": 92.1405e-3,
                "pressure_crit": 41e5,
                "temperature_crit": 591.8,
                "cp_mol_ig_comp_coeff": {
                    "A": -2.435e1,
                    "B": 5.125e-1,
                    "C": -2.765e-4,
                    "D": 4.911e-8,
                },
                "cp_mol_liq_comp_coeff": {
                    "1": 1.40e2,
                    "2": -1.52e-1,
                    "3": 6.95e-4,
                    "4": 0,
                    "5": 0,
                },
                "enth_mol_form_liq_comp_ref": 12.0e3,
                "enth_mol_form_vap_comp_ref": 50.1e3,
                "pressure_sat_comp_coeff": {
                    "A": -7.28607,
                    "B": 1.38091,
                    "C": -2.83433,
                    "D": -2.79168,
                },
            },
        },
        "C": {
            "type": Component,
            "henry_component": {"Liq": {"method": ConstantH, "type": HenryType.Kpx}},
            "enth_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            "parameter_data": {
                "mw": 92.1405e-3,
                "pressure_crit": 41e5,
                "temperature_crit": 591.8,
                "henry_ref": {"Liq": 2e5},
                "cp_mol_ig_comp_coeff": {
                    "A": -2.435e1,
                    "B": 5.125e-1,
                    "C": -2.765e-4,
                    "D": 4.911e-8,
                },
                "cp_mol_liq_comp_coeff": {
                    "1": 1.40e2,
                    "2": -1.52e-1,
                    "3": 6.95e-4,
                    "4": 0,
                    "5": 0,
                },
                "enth_mol_form_liq_comp_ref": 12.0e3,
                "enth_mol_form_vap_comp_ref": 50.1e3,
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {"type": LiquidPhase, "equation_of_state": Ideal},
        "Vap": {"type": VaporPhase, "equation_of_state": Ideal},
    },
    # Declare a base units dict to save code later
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000),
        "temperature": (273.15, 300, 450),
        "pressure": (5e4, 1e5, 1e6),
    },
    "pressure_ref": 1e5,
    "temperature_ref": 300,
    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": IdealBubbleDew,
}


class TestHenryComps0(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration)

        model.props = model.params.build_state_block([1], defined_state=True)

        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(368)
        model.props[1].pressure.fix(101325)
        model.props[1].mole_frac_comp["A"].fix(0.5)
        model.props[1].mole_frac_comp["B"].fix(0.5)
        model.props[1].mole_frac_comp["C"].fix(1e-10)

        # Trigger construction of some things we will test later
        model.props[1].pressure_bubble
        model.props[1].pressure_dew

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 2
        for i in model.params.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.params.Liq.is_liquid_phase()
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 3
        for i in model.params.component_list:
            assert i in ["A", "B", "C"]
            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 6
        for i in model.params._phase_component_set:
            assert i in [
                ("Liq", "A"),
                ("Liq", "B"),
                ("Liq", "C"),
                ("Vap", "A"),
                ("Vap", "B"),
                ("Vap", "C"),
            ]

        assert model.params.config.state_definition == FTPx

        assert model.params.config.state_bounds == {
            "flow_mol": (0, 100, 1000),
            "temperature": (273.15, 300, 450),
            "pressure": (5e4, 1e5, 1e6),
        }

        assert model.params.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE
        }

        assert isinstance(model.params.phase_equilibrium_idx, Set)
        assert len(model.params.phase_equilibrium_idx) == 3
        for i in model.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2", "PE3"]

        assert model.params.phase_equilibrium_list == {
            "PE1": ["A", ("Vap", "Liq")],
            "PE2": ["B", ("Vap", "Liq")],
            "PE3": ["C", ("Vap", "Liq")],
        }

        assert model.params.pressure_ref.value == 1e5
        assert model.params.temperature_ref.value == 300
        assert model.params.C.henry_ref_Liq.value == 2e5

    @pytest.mark.unit
    def test_init_bubble_temperature(self, model):
        _init_Tbub(model.props[1], pyunits.K)

        assert pytest.approx(365.35, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.7137, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2863, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")]
        )

    @pytest.mark.unit
    def test_init_dew_temperature(self, model):
        _init_Tdew(model.props[1], pyunits.K)

        assert pytest.approx(372.02, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2909, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.7091, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")]
        )

    @pytest.mark.unit
    def test_init_bubble_pressure(self, model):
        _init_Pbub(model.props[1])

        assert pytest.approx(109479, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")]
        )

    @pytest.mark.unit
    def test_init_dew_pressure(self, model):
        _init_Pdew(model.props[1])

        assert pytest.approx(89820, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")]
        )

    @pytest.mark.component
    def test_solve_vle(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

        assert pytest.approx(365.35, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.7137, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2863, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")]
        )

        assert pytest.approx(372.02, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2909, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.7091, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")]
        )

        assert pytest.approx(109479, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")]
        )

        assert pytest.approx(89820, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")]
        )


class TestHenryComps(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration)

        model.props = model.params.build_state_block([1], defined_state=True)

        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(368)
        model.props[1].pressure.fix(101325)
        model.props[1].mole_frac_comp["A"].fix(0.45)
        model.props[1].mole_frac_comp["B"].fix(0.45)
        model.props[1].mole_frac_comp["C"].fix(0.1)

        # Trigger construction of some things we will test later
        model.props[1].pressure_bubble
        model.props[1].pressure_dew

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 2
        for i in model.params.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.params.Liq.is_liquid_phase()
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 3
        for i in model.params.component_list:
            assert i in ["A", "B", "C"]
            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 6
        for i in model.params._phase_component_set:
            assert i in [
                ("Liq", "A"),
                ("Liq", "B"),
                ("Liq", "C"),
                ("Vap", "A"),
                ("Vap", "B"),
                ("Vap", "C"),
            ]

        assert model.params.config.state_definition == FTPx

        assert model.params.config.state_bounds == {
            "flow_mol": (0, 100, 1000),
            "temperature": (273.15, 300, 450),
            "pressure": (5e4, 1e5, 1e6),
        }

        assert model.params.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE
        }

        assert isinstance(model.params.phase_equilibrium_idx, Set)
        assert len(model.params.phase_equilibrium_idx) == 3
        for i in model.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2", "PE3"]

        assert model.params.phase_equilibrium_list == {
            "PE1": ["A", ("Vap", "Liq")],
            "PE2": ["B", ("Vap", "Liq")],
            "PE3": ["C", ("Vap", "Liq")],
        }

        assert model.params.pressure_ref.value == 1e5
        assert model.params.temperature_ref.value == 300
        assert model.params.C.henry_ref_Liq.value == 2e5

    @pytest.mark.unit
    def test_init_bubble_temperature(self, model):
        _init_Tbub(model.props[1], pyunits.K)

        assert pytest.approx(361.50, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.5750, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2276, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")]
        )
        assert pytest.approx(0.1974, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "C")]
        )

    @pytest.mark.unit
    def test_init_dew_temperature(self, model):
        _init_Tdew(model.props[1], pyunits.K)

        assert pytest.approx(370.23, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2750, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.6744, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")]
        )
        assert pytest.approx(0.0506, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "C")]
        )

    @pytest.mark.unit
    def test_init_bubble_pressure(self, model):
        _init_Pbub(model.props[1])

        assert pytest.approx(118531, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.5918, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2395, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")]
        )
        assert pytest.approx(0.1687, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "C")]
        )

    @pytest.mark.unit
    def test_init_dew_pressure(self, model):
        _init_Pdew(model.props[1])

        assert pytest.approx(95056, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2744, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.6780, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")]
        )
        assert pytest.approx(0.0476, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "C")]
        )

    @pytest.mark.component
    def test_solve_vle(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

        assert pytest.approx(361.50, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.5750, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2276, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")]
        )
        assert pytest.approx(0.1974, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "C")]
        )

        assert pytest.approx(370.23, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2750, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.6744, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")]
        )
        assert pytest.approx(0.0506, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "C")]
        )

        assert pytest.approx(118531, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")]
        )
        assert pytest.approx(0.5918, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.2395, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")]
        )
        assert pytest.approx(0.1687, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "C")]
        )

        assert pytest.approx(95056, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")]
        )
        assert pytest.approx(0.2744, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")]
        )
        assert pytest.approx(0.6780, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")]
        )
        assert pytest.approx(0.0476, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "C")]
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Hydrocarbon processing vapor phase package using Peng-Robinson EoS.

Example property package using the Generic Property Package Framework.
This example shows how to set up a property package to do hydrocarbon
processing phase equilibrium in the generic framework using Peng-Robinson
equation along with methods drawn from the pre-built IDAES property libraries.

The example includes the dictionary named configuration contains parameters
for calculating vapor phase properties for hydrocarbon processing.
"""

# Import Python libraries
import logging

# Import Pyomo units
from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import VaporPhase, Component, PhaseType as PT
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.phase_equil.forms import log_fugacity

from idaes.models.properties.modular_properties.pure import Perrys
from idaes.models.properties.modular_properties.pure import RPP4
from idaes.models.properties.modular_properties.pure import RPP5

# Set up logger
_log = logging.getLogger(__name__)


# ---------------------------------------------------------------------
# Configuration dictionary for a Prng Robinson alkene system

# Data Sources:
# [1] The Properties of Gases and Liquids (1987)
#     4th edition, Chemical Engineering Series - Robert C. Reid
# [2] Perry's Chemical Engineers' Handbook 7th Ed.
#     Converted to J/mol.K, mol/m^3
# [3] Engineering Toolbox, https://www.engineeringtoolbox.com
#     Retrieved 15th september, 2020
# [4] The Properties of Gases and Liquids (2001)
#     5th edition, Chemical Engineering Series - Robert C. Reid

configuration_vap = {
    # Specifying components
    "components": {
        "hydrogen": {
            "type": Component,
            "elemental_composition": {"H": 2, "C": 0},
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "valid_phase_types": PT.vaporPhase,
            "parameter_data": {
                "mw": (2.016e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (12.9e5, pyunits.Pa),  # [1]
                "temperature_crit": (33.2, pyunits.K),  # [1]
                "omega": -0.218,
                "cp_mol_ig_comp_coeff": {
                    "A": (2.714e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (9.274e-3, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-1.981e-5, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (7.645e-9, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "entr_mol_form_vap_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (0.0, pyunits.J / pyunits.mol),
            },
        },
        "methane": {
            "type": Component,
            "elemental_composition": {"H": 4, "C": 1},
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "valid_phase_types": PT.vaporPhase,
            "parameter_data": {
                "mw": (16.043e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (46e5, pyunits.Pa),  # [1]
                "temperature_crit": (190.4, pyunits.K),  # [1]
                "omega": 0.011,
                "cp_mol_ig_comp_coeff": {
                    "A": (1.925e1, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (5.213e-2, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (1.197e-5, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (-1.132e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "entr_mol_form_vap_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (-74600, pyunits.J / pyunits.mol),
            },
        },
        "ethane": {
            "type": Component,
            "elemental_composition": {"H": 6, "C": 2},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "enth_mol_liq_comp": Perrys,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP5,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (30.070e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (48.8e5, pyunits.Pa),  # [1]
                "temperature_crit": (305.4, pyunits.K),  # [1]
                "omega": 0.099,
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (1.9122, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.27937, None),
                    "3": (305.32, pyunits.K),
                    "4": (0.29187, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (5.409e0, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (1.781e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-6.938e-5, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (8.713e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        90.9,
                        pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
                    ),  # [2]
                    "2": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-2),
                    "3": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-3),
                    "4": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-4),
                    "5": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-5),
                },
                "enth_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_vap_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (-83800, pyunits.J / pyunits.mol),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (3.95405, None),  # [4]
                    "B": (663.720, pyunits.K),
                    "C": (256.681, pyunits.K),
                },
            },
        },
        "propane": {
            "type": Component,
            "elemental_composition": {"H": 8, "C": 3},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "enth_mol_liq_comp": Perrys,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP5,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (44.094e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (42.5e5, pyunits.Pa),  # [1]
                "temperature_crit": (369.8, pyunits.K),  # [1]
                "omega": 0.144,
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (1.3757, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.274253, None),
                    "3": (369.83, pyunits.K),
                    "4": (0.29147, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (-4.224e0, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (3.063e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-1.586e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (3.215e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        112.71,
                        pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
                    ),  # [2]
                    "2": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-2),
                    "3": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-3),
                    "4": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-4),
                    "5": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-5),
                },
                "enth_mol_form_liq_comp_ref": (0.0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_vap_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (-104700, pyunits.J / pyunits.mol),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (3.92828, None),  # [4]
                    "B": (803.9970, pyunits.K),
                    "C": (247.040, pyunits.K),
                },
            },
        },
        "nbutane": {
            "type": Component,
            "elemental_composition": {"H": 10, "C": 4},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "enth_mol_liq_comp": Perrys,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP5,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (58.124e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (38.0e5, pyunits.Pa),  # [1]
                "temperature_crit": (425.2, pyunits.K),  # [1]
                "omega": 0.199,
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (1.0677, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.27188, None),
                    "3": (425.12, pyunits.K),
                    "4": (0.28688, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (9.487e0, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (3.313e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-1.108e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (-2.822e-9, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        132.42,
                        pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
                    ),  # [2]
                    "2": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-2),
                    "3": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-3),
                    "4": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-4),
                    "5": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-5),
                },
                "enth_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_vap_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (-125600, pyunits.J / pyunits.mol),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (3.93266, None),  # [4]
                    "B": (935.7730, pyunits.K),
                    "C": (238.789, pyunits.K),
                },
            },
        },
        "ibutane": {
            "type": Component,
            "elemental_composition": {"H": 10, "C": 4},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "entr_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP5,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (58.124e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (36.5e5, pyunits.Pa),  # [1]
                "temperature_crit": (408.2, pyunits.K),  # [1]
                "omega": 0.183,
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (1.0463, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.27294, None),
                    "3": (408.14, pyunits.K),
                    "4": (0.27301, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (-1.390e0, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (3.847e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-1.846e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (2.895e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        1.7237e5,
                        pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
                    ),  # [2]
                    "2": (-1.7839e3, pyunits.J * pyunits.kmol**-1 * pyunits.K**-2),
                    "3": (1.4759e1, pyunits.J * pyunits.kmol**-1 * pyunits.K**-3),
                    "4": (-4.7909e-2, pyunits.J * pyunits.kmol**-1 * pyunits.K**-4),
                    "5": (5.8050e-5, pyunits.J * pyunits.kmol**-1 * pyunits.K**-5),
                },
                "enth_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (-135600, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    310.1,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (4.00272, None),  # [4]
                    "B": (947.5400, pyunits.K),
                    "C": (248.870, pyunits.K),
                },
            },
        },
        "ethylene": {
            "type": Component,
            "elemental_composition": {"H": 4, "C": 2},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "entr_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP5,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (28.054e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (50.5e5, pyunits.Pa),  # [1]
                "temperature_crit": (282.4, pyunits.K),  # [1]
                "omega": 0.089,
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (2.0961, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.27657, None),
                    "3": (282.34, pyunits.K),
                    "4": (0.29147, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (3.806e0, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (1.566e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-8.348e-5, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (1.755e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        2.4739e5,
                        pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
                    ),  # [2]
                    "2": (-4.4280e3, pyunits.J * pyunits.kmol**-1 * pyunits.K**-2),
                    "3": (4.0936e1, pyunits.J * pyunits.kmol**-1 * pyunits.K**-3),
                    "4": (-1.6970e-1, pyunits.J * pyunits.kmol**-1 * pyunits.K**-4),
                    "5": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-5),
                },
                "enth_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (
                    52283.264,
                    pyunits.J / pyunits.mol,
                ),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    219.5,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (3.67374, None),  # [4]
                    "B": (528.6700, pyunits.K),
                    "C": (228.790, pyunits.K),
                },
            },
        },
        "propene": {
            "type": Component,
            "elemental_composition": {"H": 6, "C": 3},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "entr_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP5,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (42.081e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (46.2e5, pyunits.Pa),  # [1]
                "temperature_crit": (365.0, pyunits.K),  # [1]
                "omega": 0.144,
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (1.4094, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.26465, None),
                    "3": (365.57, pyunits.K),
                    "4": (0.295, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (3.710e0, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (2.345e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-1.160e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (2.205e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        7.1720e5,
                        pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
                    ),  # [2]
                    "2": (-3.8632e2, pyunits.J * pyunits.kmol**-1 * pyunits.K**-2),
                    "3": (1.2348e0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-3),
                    "4": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-4),
                    "5": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-5),
                },
                "enth_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (
                    20413.736,
                    pyunits.J / pyunits.mol,
                ),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    266.9,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (3.95606, None),  # [4]
                    "B": (789.6200, pyunits.K),
                    "C": (247.580, pyunits.K),
                },
            },
        },
        "butene": {
            "type": Component,
            "elemental_composition": {"H": 8, "C": 4},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "entr_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP5,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (56.104e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (40.2e5, pyunits.Pa),  # [1]
                "temperature_crit": (419.3, pyunits.K),  # [1]
                "omega": 0.191,
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (1.0972, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.2649, None),
                    "3": (419.95, pyunits.K),
                    "4": (0.29043, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (-2.994e0, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (3.532e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-1.990e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (4.463e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        1.3589e5,
                        pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
                    ),  # [2]
                    "2": (-4.7739e2, pyunits.J * pyunits.kmol**-1 * pyunits.K**-2),
                    "3": (2.1835e0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-3),
                    "4": (-2.2230e-3, pyunits.J * pyunits.kmol**-1 * pyunits.K**-4),
                    "5": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-5),
                },
                "enth_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (1171.52, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    305.6,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (3.96640, None),  # [4]
                    "B": (927.2100, pyunits.K),
                    "C": (238.630, pyunits.K),
                },
            },
        },
        "pentene": {
            "type": Component,
            "elemental_composition": {"H": 10, "C": 5},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "entr_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP5,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (70.135e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (40.5e5, pyunits.Pa),  # [1]
                "temperature_crit": (464.7, pyunits.K),  # [1]
                "omega": 0.233,  # [1]
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (0.9038, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.26648, None),
                    "3": (464.78, pyunits.K),
                    "4": (0.2905, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (-1.340e-1, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (4.329e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-2.317e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (4.681e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        1.5467e5,
                        pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
                    ),  # [2]
                    "2": (-4.2600e2, pyunits.J * pyunits.kmol**-1 * pyunits.K**-2),
                    "3": (1.964, pyunits.J * pyunits.kmol**-1 * pyunits.K**-3),
                    "4": (-1.8034e-3, pyunits.J * pyunits.kmol**-1 * pyunits.K**-4),
                    "5": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-5),
                },
                "enth_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (
                    -20920.00,
                    pyunits.J / pyunits.mol,
                ),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    347.82,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (3.96914, None),  # [4]
                    "B": (1044.010, pyunits.K),
                    "C": (233.450, pyunits.K),
                },
            },
        },
        "hexene": {
            "type": Component,
            "elemental_composition": {"H": 12, "C": 6},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "entr_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP5,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (84.162e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (31.7e5, pyunits.Pa),  # [1]
                "temperature_crit": (504.0, pyunits.K),  # [1]
                "omega": 0.285,  # [1]
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (0.7389, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.26147, None),
                    "3": (504.03, pyunits.K),
                    "4": (0.2902, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (-1.746e0, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (5.309e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-2.903e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (6.054e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        1.9263e5,
                        pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
                    ),  # [2]
                    "2": (-5.7116e2, pyunits.J * pyunits.kmol**-1 * pyunits.K**-2),
                    "3": (2.4004e0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-3),
                    "4": (-1.9758e-3, pyunits.J * pyunits.kmol**-1 * pyunits.K**-4),
                    "5": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-5),
                },
                "enth_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (
                    -41672.64,
                    pyunits.J / pyunits.mol,
                ),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    385.97,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (3.98260, None),  # [4]
                    "B": (1148.620, pyunits.K),
                    "C": (225.340, pyunits.K),
                },
            },
        },
        "heptene": {
            "type": Component,
            "elemental_composition": {"H": 14, "C": 7},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "entr_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP5,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (98.189e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (25.4e5, pyunits.Pa),  # [1]
                "temperature_crit": (537.2, pyunits.K),  # [1]
                "omega": 0.358,  # [1]
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (0.63734, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.26319, None),
                    "3": (537.29, pyunits.K),
                    "4": (0.27375, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (-3.303e0, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (6.297e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-3.512e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (7.607e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        1.8997e5,
                        pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
                    ),  # [2]
                    "2": (-1.5670e2, pyunits.J * pyunits.kmol**-1 * pyunits.K**-2),
                    "3": (3.4300e-1, pyunits.J * pyunits.kmol**-1 * pyunits.K**-3),
                    "4": (1.5222e-3, pyunits.J * pyunits.kmol**-1 * pyunits.K**-4),
                    "5": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-5),
                },
                "enth_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (
                    -62299.76,
                    pyunits.J / pyunits.mol,
                ),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    424,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (4.02677, None),  # [4]
                    "B": (1258.340, pyunits.K),
                    "C": (219.300, pyunits.K),
                },
            },
        },
        "octene": {
            "type": Component,
            "elemental_composition": {"H": 16, "C": 8},
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "entr_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP5,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (112.216e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (26.2e5, pyunits.Pa),  # [1]
                "temperature_crit": (566.6, pyunits.K),  # [1]
                "omega": 0.386,  # [1]
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (0.5871, pyunits.kmol * pyunits.m**-3),  # [2] pg. 2-98
                    "2": (0.27005, None),
                    "3": (566.65, pyunits.K),
                    "4": (0.27187, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (-4.099e0, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (7.239e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-4.036e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (8.675e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        3.7930e5,
                        pyunits.J * pyunits.kmol**-1 * pyunits.K**-1,
                    ),  # [2]
                    "2": (-2.1175e3, pyunits.J * pyunits.kmol**-1 * pyunits.K**-2),
                    "3": (8.2362e0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-3),
                    "4": (-9.0093e-3, pyunits.J * pyunits.kmol**-1 * pyunits.K**-4),
                    "5": (0, pyunits.J * pyunits.kmol**-1 * pyunits.K**-5),
                },
                "enth_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_liq_comp_ref": (0, pyunits.J / pyunits.mol / pyunits.K),
                "enth_mol_form_vap_comp_ref": (
                    -82926.88,
                    pyunits.J / pyunits.mol,
                ),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    454.36,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (4.05985, None),  # [4]
                    "B": (1355.460, pyunits.K),
                    "C": (213.050, pyunits.K),
                },
            },
        },
    },
    # Specifying phases
    "phases": {
        "Vap": {
            "type": VaporPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        }
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (273.15, 300, 1500, pyunits.K),
        "pressure": (5e4, 1e5, 1e7, pyunits.Pa),
    },
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),
    # Defining phase equilibria
    "parameter_data": {
        "PR_kappa": {
            ("hydrogen", "hydrogen"): 0.000,
            ("hydrogen", "methane"): 0.000,
            ("hydrogen", "ethane"): 0.000,
            ("hydrogen", "propane"): 0.000,
            ("hydrogen", "nbutane"): 0.000,
            ("hydrogen", "ibutane"): 0.000,
            ("hydrogen", "ethylene"): 0.000,
            ("hydrogen", "propene"): 0.000,
            ("hydrogen", "butene"): 0.000,
            ("hydrogen", "pentene"): 0.000,
            ("hydrogen", "hexene"): 0.000,
            ("hydrogen", "heptene"): 0.000,
            ("hydrogen", "octene"): 0.000,
            ("methane", "hydrogen"): 0.000,
            ("methane", "methane"): 0.000,
            ("methane", "ethane"): 0.000,
            ("methane", "propane"): 0.000,
            ("methane", "nbutane"): 0.000,
            ("methane", "ibutane"): 0.000,
            ("methane", "ethylene"): 0.000,
            ("methane", "propene"): 0.000,
            ("methane", "butene"): 0.000,
            ("methane", "pentene"): 0.000,
            ("methane", "hexene"): 0.000,
            ("methane", "heptene"): 0.000,
            ("methane", "octene"): 0.000,
            ("ethane", "hydrogen"): 0.000,
            ("ethane", "methane"): 0.000,
            ("ethane", "ethane"): 0.000,
            ("ethane", "propane"): 0.000,
            ("ethane", "nbutane"): 0.000,
            ("ethane", "ibutane"): 0.000,
            ("ethane", "ethylene"): 0.000,
            ("ethane", "propene"): 0.000,
            ("ethane", "butene"): 0.000,
            ("ethane", "pentene"): 0.000,
            ("ethane", "hexene"): 0.000,
            ("ethane", "heptene"): 0.000,
            ("ethane", "octene"): 0.000,
            ("propane", "hydrogen"): 0.000,
            ("propane", "methane"): 0.000,
            ("propane", "ethane"): 0.000,
            ("propane", "propane"): 0.000,
            ("propane", "nbutane"): 0.000,
            ("propane", "ibutane"): 0.000,
            ("propane", "ethylene"): 0.000,
            ("propane", "propene"): 0.000,
            ("propane", "butene"): 0.000,
            ("propane", "pentene"): 0.000,
            ("propane", "hexene"): 0.000,
            ("propane", "heptene"): 0.000,
            ("propane", "octene"): 0.000,
            ("nbutane", "hydrogen"): 0.000,
            ("nbutane", "methane"): 0.000,
            ("nbutane", "ethane"): 0.000,
            ("nbutane", "propane"): 0.000,
            ("nbutane", "nbutane"): 0.000,
            ("nbutane", "ibutane"): 0.000,
            ("nbutane", "ethylene"): 0.000,
            ("nbutane", "propene"): 0.000,
            ("nbutane", "butene"): 0.000,
            ("nbutane", "pentene"): 0.000,
            ("nbutane", "hexene"): 0.000,
            ("nbutane", "heptene"): 0.000,
            ("nbutane", "octene"): 0.000,
            ("ibutane", "hydrogen"): 0.000,
            ("ibutane", "methane"): 0.000,
            ("ibutane", "ethane"): 0.000,
            ("ibutane", "propane"): 0.000,
            ("ibutane", "nbutane"): 0.000,
            ("ibutane", "ibutane"): 0.000,
            ("ibutane", "ethylene"): 0.000,
            ("ibutane", "propene"): 0.000,
            ("ibutane", "butene"): 0.000,
            ("ibutane", "pentene"): 0.000,
            ("ibutane", "hexene"): 0.000,
            ("ibutane", "heptene"): 0.000,
            ("ibutane", "octene"): 0.000,
            ("ethylene", "hydrogen"): 0.000,
            ("ethylene", "methane"): 0.000,
            ("ethylene", "ethane"): 0.000,
            ("ethylene", "propane"): 0.000,
            ("ethylene", "nbutane"): 0.000,
            ("ethylene", "ibutane"): 0.000,
            ("ethylene", "ethylene"): 0.000,
            ("ethylene", "propene"): 0.000,
            ("ethylene", "butene"): 0.000,
            ("ethylene", "pentene"): 0.000,
            ("ethylene", "hexene"): 0.000,
            ("ethylene", "heptene"): 0.000,
            ("ethylene", "octene"): 0.000,
            ("propene", "hydrogen"): 0.000,
            ("propene", "methane"): 0.000,
            ("propene", "ethane"): 0.000,
            ("propene", "propane"): 0.000,
            ("propene", "nbutane"): 0.000,
            ("propene", "ibutane"): 0.000,
            ("propene", "ethylene"): 0.000,
            ("propene", "propene"): 0.000,
            ("propene", "butene"): 0.000,
            ("propene", "pentene"): 0.000,
            ("propene", "hexene"): 0.000,
            ("propene", "heptene"): 0.000,
            ("propene", "octene"): 0.000,
            ("butene", "hydrogen"): 0.000,
            ("butene", "methane"): 0.000,
            ("butene", "ethane"): 0.000,
            ("butene", "propane"): 0.000,
            ("butene", "nbutane"): 0.000,
            ("butene", "ibutane"): 0.000,
            ("butene", "ethylene"): 0.000,
            ("butene", "propene"): 0.000,
            ("butene", "butene"): 0.000,
            ("butene", "pentene"): 0.000,
            ("butene", "hexene"): 0.000,
            ("butene", "heptene"): 0.000,
            ("butene", "octene"): 0.000,
            ("pentene", "hydrogen"): 0.000,
            ("pentene", "methane"): 0.000,
            ("pentene", "ethane"): 0.000,
            ("pentene", "propane"): 0.000,
            ("pentene", "nbutane"): 0.000,
            ("pentene", "ibutane"): 0.000,
            ("pentene", "ethylene"): 0.000,
            ("pentene", "propene"): 0.000,
            ("pentene", "butene"): 0.000,
            ("pentene", "pentene"): 0.000,
            ("pentene", "hexene"): 0.000,
            ("pentene", "heptene"): 0.000,
            ("pentene", "octene"): 0.000,
            ("hexene", "hydrogen"): 0.000,
            ("hexene", "methane"): 0.000,
            ("hexene", "ethane"): 0.000,
            ("hexene", "propane"): 0.000,
            ("hexene", "nbutane"): 0.000,
            ("hexene", "ibutane"): 0.000,
            ("hexene", "ethylene"): 0.000,
            ("hexene", "propene"): 0.000,
            ("hexene", "butene"): 0.000,
            ("hexene", "pentene"): 0.000,
            ("hexene", "hexene"): 0.000,
            ("hexene", "heptene"): 0.000,
            ("hexene", "octene"): 0.000,
            ("heptene", "hydrogen"): 0.000,
            ("heptene", "methane"): 0.000,
            ("heptene", "ethane"): 0.000,
            ("heptene", "propane"): 0.000,
            ("heptene", "nbutane"): 0.000,
            ("heptene", "ibutane"): 0.000,
            ("heptene", "ethylene"): 0.000,
            ("heptene", "propene"): 0.000,
            ("heptene", "butene"): 0.000,
            ("heptene", "pentene"): 0.000,
            ("heptene", "hexene"): 0.000,
            ("heptene", "heptene"): 0.000,
            ("heptene", "octene"): 0.000,
            ("octene", "hydrogen"): 0.000,
            ("octene", "methane"): 0.000,
            ("octene", "ethane"): 0.000,
            ("octene", "propane"): 0.000,
            ("octene", "nbutane"): 0.000,
            ("octene", "ibutane"): 0.000,
            ("octene", "ethylene"): 0.000,
            ("octene", "propene"): 0.000,
            ("octene", "butene"): 0.000,
            ("octene", "pentene"): 0.000,
            ("octene", "hexene"): 0.000,
            ("octene", "heptene"): 0.000,
            ("octene", "octene"): 0.000,
        }
    },
}
